# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/dialogs/choose_format_device.ui'
#
# Created by: PyQt6 UI code generator 6.4.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_ChooseFormatDeviceDialog(object):
    def setupUi(self, ChooseFormatDeviceDialog):
        ChooseFormatDeviceDialog.setObjectName("ChooseFormatDeviceDialog")
        ChooseFormatDeviceDialog.resize(507, 377)
        icon = QtGui.QIcon.ic("mimetypes/unknown.png")
        ChooseFormatDeviceDialog.setWindowIcon(icon)
        self.vboxlayout = QtWidgets.QVBoxLayout(ChooseFormatDeviceDialog)
        self.vboxlayout.setObjectName("vboxlayout")
        self.msg = QtWidgets.QLabel(ChooseFormatDeviceDialog)
        self.msg.setText("")
        self.msg.setObjectName("msg")
        self.vboxlayout.addWidget(self.msg)
        self.formats = QtWidgets.QTreeWidget(ChooseFormatDeviceDialog)
        self.formats.setAlternatingRowColors(True)
        self.formats.setIconSize(QtCore.QSize(64, 64))
        self.formats.setAllColumnsShowFocus(True)
        self.formats.setObjectName("formats")
        self.formats.headerItem().setTextAlignment(1, QtCore.Qt.AlignmentFlag.AlignLeft|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.vboxlayout.addWidget(self.formats)
        self.buttonBox = QtWidgets.QDialogButtonBox(ChooseFormatDeviceDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.vboxlayout.addWidget(self.buttonBox)

        self.retranslateUi(ChooseFormatDeviceDialog)
        self.buttonBox.accepted.connect(ChooseFormatDeviceDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(ChooseFormatDeviceDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(ChooseFormatDeviceDialog)

    def retranslateUi(self, ChooseFormatDeviceDialog):

        ChooseFormatDeviceDialog.setWindowTitle(_("Choose format"))
        self.formats.headerItem().setText(0, _("Format"))
        self.formats.headerItem().setText(1, _("Existing"))
        self.formats.headerItem().setText(2, _("Convertible"))
