#!/usr/bin/env python


__license__   = 'GPL v3'
__copyright__ = '2012, Kovid Goyal <kovid at kovidgoyal.net>'
__docformat__ = 'restructuredtext en'


class UnsupportedFont(ValueError):
    pass


class NoGlyphs(ValueError):
    pass

