# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/preferences/toolbar.ui'
#
# Created by: PyQt6 UI code generator 6.4.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(831, 553)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(Form)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.what = QtWidgets.QComboBox(Form)
        self.what.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.what.setMinimumContentsLength(20)
        self.what.setObjectName("what")
        self.verticalLayout_2.addWidget(self.what)
        self.actions_widget = QtWidgets.QWidget(Form)
        self.actions_widget.setObjectName("actions_widget")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.actions_widget)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.groupBox = QtWidgets.QGroupBox(self.actions_widget)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.all_actions = QtWidgets.QListView(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(100)
        sizePolicy.setHeightForWidth(self.all_actions.sizePolicy().hasHeightForWidth())
        self.all_actions.setSizePolicy(sizePolicy)
        self.all_actions.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.all_actions.setIconSize(QtCore.QSize(32, 32))
        self.all_actions.setWordWrap(True)
        self.all_actions.setObjectName("all_actions")
        self.verticalLayout.addWidget(self.all_actions)
        self.horizontalLayout_2.addWidget(self.groupBox)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.add_action_button = QtWidgets.QToolButton(self.actions_widget)
        icon = QtGui.QIcon.ic("forward.png")
        self.add_action_button.setIcon(icon)
        self.add_action_button.setIconSize(QtCore.QSize(24, 24))
        self.add_action_button.setObjectName("add_action_button")
        self.verticalLayout_3.addWidget(self.add_action_button)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
        self.verticalLayout_3.addItem(spacerItem)
        self.remove_action_button = QtWidgets.QToolButton(self.actions_widget)
        icon1 = QtGui.QIcon.ic("back.png")
        self.remove_action_button.setIcon(icon1)
        self.remove_action_button.setIconSize(QtCore.QSize(24, 24))
        self.remove_action_button.setObjectName("remove_action_button")
        self.verticalLayout_3.addWidget(self.remove_action_button)
        self.horizontalLayout_2.addLayout(self.verticalLayout_3)
        self.groupBox_2 = QtWidgets.QGroupBox(self.actions_widget)
        self.groupBox_2.setObjectName("groupBox_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.groupBox_2)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.current_actions = QtWidgets.QListView(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(100)
        sizePolicy.setHeightForWidth(self.current_actions.sizePolicy().hasHeightForWidth())
        self.current_actions.setSizePolicy(sizePolicy)
        self.current_actions.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.current_actions.setIconSize(QtCore.QSize(32, 32))
        self.current_actions.setWordWrap(True)
        self.current_actions.setObjectName("current_actions")
        self.horizontalLayout.addWidget(self.current_actions)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.action_up_button = QtWidgets.QToolButton(self.groupBox_2)
        icon2 = QtGui.QIcon.ic("arrow-up.png")
        self.action_up_button.setIcon(icon2)
        self.action_up_button.setIconSize(QtCore.QSize(24, 24))
        self.action_up_button.setObjectName("action_up_button")
        self.verticalLayout_4.addWidget(self.action_up_button)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_4.addItem(spacerItem1)
        self.action_down_button = QtWidgets.QToolButton(self.groupBox_2)
        icon3 = QtGui.QIcon.ic("arrow-down.png")
        self.action_down_button.setIcon(icon3)
        self.action_down_button.setIconSize(QtCore.QSize(24, 24))
        self.action_down_button.setObjectName("action_down_button")
        self.verticalLayout_4.addWidget(self.action_down_button)
        self.horizontalLayout.addLayout(self.verticalLayout_4)
        self.horizontalLayout_2.addWidget(self.groupBox_2)
        self.verticalLayout_2.addWidget(self.actions_widget)
        self.help_text = QtWidgets.QLabel(Form)
        self.help_text.setText("")
        self.help_text.setWordWrap(True)
        self.help_text.setObjectName("help_text")
        self.verticalLayout_2.addWidget(self.help_text)
        self.spacer_widget = QtWidgets.QWidget(Form)
        self.spacer_widget.setObjectName("spacer_widget")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.spacer_widget)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        spacerItem2 = QtWidgets.QSpacerItem(20, 224, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_5.addItem(spacerItem2)
        self.verticalLayout_2.addWidget(self.spacer_widget)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):

        Form.setWindowTitle(_("Form"))
        self.label.setText(_("<p>The toolbar in calibre is different depending on whether a device is connected or not. Choose <b>which toolbar</b> you would like to customize:"))
        self.what.setToolTip(_("Choose the toolbar to customize"))
        self.groupBox.setTitle(_("A&vailable actions"))
        self.add_action_button.setToolTip(_("Add selected actions to toolbar"))
        self.add_action_button.setText(_("..."))
        self.remove_action_button.setToolTip(_("Remove selected actions from toolbar"))
        self.remove_action_button.setText(_("..."))
        self.groupBox_2.setTitle(_("&Current actions"))
        self.action_up_button.setToolTip(_("Move selected action up"))
        self.action_up_button.setText(_("..."))
        self.action_down_button.setToolTip(_("Move selected action down"))
        self.action_down_button.setText(_("..."))
        self.action_down_button.setShortcut(_("Ctrl+S"))
