# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/dialogs/edit_authors_dialog.ui'
#
# Created by: PyQt6 UI code generator 6.7.0.dev2404081550
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_EditAuthorsDialog(object):
    def setupUi(self, EditAuthorsDialog):
        EditAuthorsDialog.setObjectName("EditAuthorsDialog")
        EditAuthorsDialog.resize(768, 342)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(EditAuthorsDialog.sizePolicy().hasHeightForWidth())
        EditAuthorsDialog.setSizePolicy(sizePolicy)
        self.verticalLayout = QtWidgets.QVBoxLayout(EditAuthorsDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridlayout = QtWidgets.QGridLayout()
        self.gridlayout.setObjectName("gridlayout")
        self.filter_box = HistoryLineEdit(parent=EditAuthorsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.filter_box.sizePolicy().hasHeightForWidth())
        self.filter_box.setSizePolicy(sizePolicy)
        self.filter_box.setMinimumSize(QtCore.QSize(200, 0))
        self.filter_box.setProperty("clearButtonEnabled", True)
        self.filter_box.setObjectName("filter_box")
        self.gridlayout.addWidget(self.filter_box, 1, 1, 1, 1)
        self.label = QtWidgets.QLabel(parent=EditAuthorsDialog)
        self.label.setObjectName("label")
        self.gridlayout.addWidget(self.label, 0, 0, 1, 1)
        self.find_box = HistoryLineEdit(parent=EditAuthorsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.find_box.sizePolicy().hasHeightForWidth())
        self.find_box.setSizePolicy(sizePolicy)
        self.find_box.setMinimumSize(QtCore.QSize(200, 0))
        self.find_box.setProperty("clearButtonEnabled", True)
        self.find_box.setObjectName("find_box")
        self.gridlayout.addWidget(self.find_box, 0, 1, 1, 1)
        self.label1 = QtWidgets.QLabel(parent=EditAuthorsDialog)
        self.label1.setObjectName("label1")
        self.gridlayout.addWidget(self.label1, 1, 0, 1, 1)
        self.filter_button = QtWidgets.QPushButton(parent=EditAuthorsDialog)
        self.filter_button.setObjectName("filter_button")
        self.gridlayout.addWidget(self.filter_button, 1, 2, 1, 1)
        self.find_button = QtWidgets.QPushButton(parent=EditAuthorsDialog)
        self.find_button.setObjectName("find_button")
        self.gridlayout.addWidget(self.find_button, 0, 2, 1, 1)
        self.show_button_layout = QtWidgets.QVBoxLayout()
        self.show_button_layout.setObjectName("show_button_layout")
        self.apply_all_checkbox = QtWidgets.QRadioButton(parent=EditAuthorsDialog)
        self.apply_all_checkbox.setObjectName("apply_all_checkbox")
        self.show_button_layout.addWidget(self.apply_all_checkbox)
        self.apply_vl_checkbox = QtWidgets.QRadioButton(parent=EditAuthorsDialog)
        self.apply_vl_checkbox.setObjectName("apply_vl_checkbox")
        self.show_button_layout.addWidget(self.apply_vl_checkbox)
        self.apply_selection_checkbox = QtWidgets.QRadioButton(parent=EditAuthorsDialog)
        self.apply_selection_checkbox.setObjectName("apply_selection_checkbox")
        self.show_button_layout.addWidget(self.apply_selection_checkbox)
        self.gridlayout.addLayout(self.show_button_layout, 0, 3, 2, 1)
        self.verticalLayout.addLayout(self.gridlayout)
        self.table = QtWidgets.QTableWidget(parent=EditAuthorsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.table.sizePolicy().hasHeightForWidth())
        self.table.setSizePolicy(sizePolicy)
        self.table.setColumnCount(0)
        self.table.setObjectName("table")
        self.table.setRowCount(0)
        self.verticalLayout.addWidget(self.table)
        self.gridlayout1 = QtWidgets.QGridLayout()
        self.gridlayout1.setObjectName("gridlayout1")
        self.recalc_author_sort = QtWidgets.QPushButton(parent=EditAuthorsDialog)
        self.recalc_author_sort.setObjectName("recalc_author_sort")
        self.gridlayout1.addWidget(self.recalc_author_sort, 1, 0, 1, 1)
        self.auth_sort_to_author = QtWidgets.QPushButton(parent=EditAuthorsDialog)
        self.auth_sort_to_author.setObjectName("auth_sort_to_author")
        self.gridlayout1.addWidget(self.auth_sort_to_author, 1, 1, 1, 1)
        self.edit_current_cell = QtWidgets.QPushButton(parent=EditAuthorsDialog)
        self.edit_current_cell.setObjectName("edit_current_cell")
        self.gridlayout1.addWidget(self.edit_current_cell, 1, 3, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=EditAuthorsDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonBox.sizePolicy().hasHeightForWidth())
        self.buttonBox.setSizePolicy(sizePolicy)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setCenterButtons(False)
        self.buttonBox.setObjectName("buttonBox")
        self.gridlayout1.addWidget(self.buttonBox, 1, 4, 1, 1)
        self.verticalLayout.addLayout(self.gridlayout1)
        self.label.setBuddy(self.find_box)
        self.label1.setBuddy(self.filter_box)

        self.retranslateUi(EditAuthorsDialog)
        self.buttonBox.accepted.connect(EditAuthorsDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(EditAuthorsDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(EditAuthorsDialog)

    def retranslateUi(self, EditAuthorsDialog):

        EditAuthorsDialog.setWindowTitle(_("Manage authors"))
        self.filter_box.setToolTip(_("<p>Only show authors that contain the text in this box.\n"
"          The match ignores case.</p>"))
        self.label.setText(_("&Search for:"))
        self.label1.setText(_("Filter &by:"))
        self.filter_button.setText(_("Fi&lter"))
        self.find_button.setText(_("S&earch"))
        self.apply_all_checkbox.setText(_("Show &all authors"))
        self.apply_vl_checkbox.setToolTip(_("<p>Show authors only if they appear in the\n"
"                    current Virtual library. Edits already done may be hidden but will\n"
"                    not be forgotten.\n"
"                    </p><p>\n"
"                    Note that this box affects only what is displayed. Changes\n"
"                    will affect all books in your library even if this box\n"
"                    is checked.</p><"))
        self.apply_vl_checkbox.setText(_("Only show authors in the current &Virtual library"))
        self.apply_selection_checkbox.setToolTip(_("<p>Show items only if they appear in the\n"
"                    currently selected books. Edits already done may be hidden but will\n"
"                    not be forgotten.\n"
"                    </p><p>\n"
"                    Note that this box affects only what is displayed. Changes\n"
"                    will affect all books in your library even if this box\n"
"                    is checked.</p><"))
        self.apply_selection_checkbox.setText(_("O&nly show authors in the currently selected books"))
        self.recalc_author_sort.setToolTip(_("Reset all the author sort values to a value automatically\n"
"generated from the author. Exactly how this value is automatically\n"
"generated can be controlled via Preferences->Advanced->Tweaks"))
        self.recalc_author_sort.setText(_("&Recalculate all author sort values"))
        self.auth_sort_to_author.setToolTip(_("Copy author sort to author for every author. You typically use this button\n"
"after changing Preferences->Advanced->Tweaks->Author sort name algorithm"))
        self.auth_sort_to_author.setText(_("&Copy all author sort values to author"))
        self.edit_current_cell.setToolTip(_("<p>Edit the currently selected cell. If an author is edited then it is renamed\n"
"                in every book where it is used. Double-clicking and pressing the edit\n"
"                key also work.</p>"))
        self.edit_current_cell.setText(_("Edi&t current cell"))
from calibre.gui2.widgets import HistoryLineEdit
